/*
** sem.c                                 interface to SysV semaphore functions
**
** version: 1.1
**
** author: Peter Eriksson, pen@lysator.liu.se
**
** history:
**   901009 pen      initial coding
**   901009 pen      added support for sem_stat and sem_access
*/

#include <stdio.h>
#include "sem.h"


/* Create a new semaphore group with default access (uid:all, rest:no) */
int sem_create(int key, int nsems)
{
  int semid;
  struct semid_ds buf;
  

  semid = semget(key, nsems, IPC_CREAT | 0600);
  
  return semid;
}


/* Set the access rights to this semaphore group */
int sem_access(int sid, int uid, int gid, int prot)
{
  struct semid_ds buf;

  buf.sem_perm.uid = uid;
  buf.sem_perm.gid = gid;
  buf.sem_perm.mode = prot;

  return semctl(sid, 0, IPC_SET, &buf);
}


/* Destroy a semaphore group */
int sem_destroy(int sid)
{
  return semctl(sid, 0, IPC_RMID, 0);
}


/* Get the value of a semaphore in a group */
int sem_get(int sid, int sno)
{
  return semctl(sid, sno, GETVAL, 0);
}


/* Set the value of a semaphore in a group */
int sem_set(int sid, int sno, int val)
{
  return semctl(sid, sno, SETVAL, val);
}


/* Wait for a semaphore value to become 0 (zero) */
int sem_wait(int sid, int sno)
{
  struct sembuf sop[1];

  sop[0].sem_num = sno;
  sop[0].sem_op  = 0;
  sop[0].sem_flg = 0;

  return semop(sid,(struct sembuf *) &sop, 1);
}


/* Add a value to a semaphore */
int sem_add(int sid, int sno, int val)
{
  struct sembuf sop[1];

  sop[0].sem_num = sno;
  sop[0].sem_op  = val;
  sop[0].sem_flg = 0;

  return semop(sid,(struct sembuf *) &sop, 1);
}


/* Subtract a value from a semaphore */
int sem_sub(int sid, int sno, int val)
{
  struct sembuf sop[1];

  sop[0].sem_num = sno;
  sop[0].sem_op  = -val;
  sop[0].sem_flg = IPC_NOWAIT;

  return semop(sid,(struct sembuf *) &sop, 1);
}


int sem_stat(int semid, struct semid_ds *buf)
{
  return semctl(semid, 0, IPC_STAT, buf);
}



#if DEBUG
main(int argc, char *argv[])
{
  int sid;
  int val;
  

  if (argc < 2)
  {
    printf("syntax: %s <key-number>\n", argv[0]);
    exit(1);
  }
  
  sid = sem_create(atoi(argv[1]), 10);
  printf("sem_create = %d\n", sid);
  
  if (sid == -1)
  {
    perror("sem_create");
    exit(1);
  }

  sem_set(sid, 0, 1); 
  
  if (fork())
  {
    /* parent */
    puts("parent: Waiting for child...");
    if (sem_wait(sid, 0) == -1)
      perror("parent: sem_wait");
    puts("parent: Semaphore raised");
  }
  else
  {
    /* child */
    puts("child: Sleeping 5 seconds");
    sleep(5);
    puts("child: Signalling...");
    if (sem_set(sid, 0, 0) == -1)
      perror("child: sem_sub");
    sleep(5);
    puts("child: Exiting.");
    exit(0);
  }
}
#endif
